@@ -43,33 +43,40 @@ enum LineMark {
     LineTested,
     LineMaybeTested,
     LineNotTested,
+    LineFlakyTested,
     BeginMaybeTested,
     BeginNotTested,
+    BeginFlakyTested,
     EndMaybeTested,
     EndNotTested,
+    EndFlakyTested,
     FileMaybeTested,
     FileNotTested,
+    FileFlakyTested,
 }
 
-#[derive(Clone)]
 #[doc(hidden)]
+#[derive(Clone, Copy, Debug)]
 enum LineAnnotation {
     Tested(bool),
     MaybeTested(bool),
     NotTested(bool),
+    FlakyTested(bool),
 }
 
 #[doc(hidden)]
-const fn is_explicit(line_annotation: &LineAnnotation) -> bool {
+const fn is_explicit(line_annotation: LineAnnotation) -> bool {
     matches!(
-        *line_annotation,
+        line_annotation,
         LineAnnotation::Tested(true)
             | LineAnnotation::MaybeTested(true)
             | LineAnnotation::NotTested(true)
+            | LineAnnotation::FlakyTested(true)
     )
 }
 
 #[doc(hidden)]
+#[derive(Debug)]
 enum FileAnnotations {
     LineAnnotations(Vec<LineAnnotation>),
     MaybeTested,
@@ -78,22 +85,25 @@ enum FileAnnotations {
 
 #[doc(hidden)]
 fn main() {
-    process_args();
+    let flaky_policy = process_args();
 
     let mut coverage_annotations = HashMap::new();
     let mut source_annotations = HashMap::new();
     collect_dir_annotations(
+        flaky_policy,
         Path::new("."),
         &mut source_annotations,
         &mut coverage_annotations,
     )
     .unwrap();
-    let exit_status = report_wrong_annotations(&coverage_annotations, &source_annotations);
+    let exit_status =
+        report_wrong_annotations(flaky_policy, &coverage_annotations, &source_annotations);
     std::process::exit(exit_status);
 }
 
 #[doc(hidden)]
 fn collect_dir_annotations(
+    flaky_policy: FlakyPolicy,
     dir: &Path,
     source_annotations: &mut HashMap<String, FileAnnotations>,
     coverage_annotations: &mut HashMap<String, HashMap<i32, bool>>,
@@ -103,13 +113,18 @@ fn collect_dir_annotations(
         let entry: fs::DirEntry = entry?;
         let path = entry.path();
         if path.is_dir() {
-            collect_dir_annotations(&path, source_annotations, coverage_annotations)?;
+            collect_dir_annotations(
+                flaky_policy,
+                &path,
+                source_annotations,
+                coverage_annotations,
+            )?;
         } else if let Ok(canonical) = fs::canonicalize(path) {
             let file_name = canonical.as_path().to_str().unwrap();
             if file_name.ends_with("/cobertura.xml") {
                 collect_coverage_annotations(canonical.as_path(), coverage_annotations);
             } else if file_name.ends_with(".rs") {
-                let annotations = collect_file_annotations(canonical.as_path());
+                let annotations = collect_file_annotations(flaky_policy, canonical.as_path());
                 source_annotations.insert(file_name.to_string(), annotations);
             }
         }
@@ -119,12 +134,13 @@ fn collect_dir_annotations(
 
 #[allow(clippy::too_many_lines)]
 #[doc(hidden)]
-fn collect_file_annotations(path: &Path) -> FileAnnotations {
+fn collect_file_annotations(flaky_policy: FlakyPolicy, path: &Path) -> FileAnnotations {
     let file = File::open(path).unwrap_or_else(|_| panic!("can't open {}", path.to_str().unwrap()));
     let file = BufReader::new(file);
     let mut region_annotation = LineAnnotation::Tested(false);
     let mut is_file_not_tested = false;
     let mut is_file_maybe_tested = false;
+    let mut is_file_flaky_tested = false;
     let mut line_annotations = Vec::new();
     let untrusted_regex =
         Regex::new(r"^(?:\s*\}(?:\s*\))*(?:\s*;)?|\s*(?:\}\s*)?else(?:\s*\{)?)?\s*(?://.*)?$")
@@ -132,7 +148,7 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
     for (mut line_number, line) in file.lines().enumerate() {
         line_number += 1;
         let line_text = line.unwrap();
-        let line_mark = extract_line_mark(line_text.as_ref());
+        let line_mark = extract_line_mark(path.to_str().unwrap(), line_number, line_text.as_ref());
         let (line_annotation, next_region_annotation) = match (line_mark, region_annotation) {
             (LineMark::None, region_annotation) => {
                 if untrusted_regex.is_match(line_text.as_ref()) {
@@ -140,7 +156,7 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
                 } else if line_text.contains("unreachable!()") {
                     (LineAnnotation::NotTested(false), region_annotation)
                 } else {
-                    (region_annotation.clone(), region_annotation)
+                    (region_annotation, region_annotation)
                 }
             }
 
@@ -186,6 +202,21 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
                 (LineAnnotation::MaybeTested(true), region_annotation)
             }
 
+            (LineMark::LineFlakyTested, LineAnnotation::FlakyTested(_)) => {
+                eprintln!(
+                    "{}:{}: redundant FLAKY TESTED coverage annotation",
+                    path.to_str().unwrap(),
+                    line_number
+                );
+                (
+                    LineAnnotation::FlakyTested(true),
+                    LineAnnotation::FlakyTested(false),
+                )
+            }
+            (LineMark::LineFlakyTested, region_annotation) => {
+                (LineAnnotation::FlakyTested(true), region_annotation)
+            }
+
             (LineMark::BeginNotTested, LineAnnotation::Tested(_)) => (
                 LineAnnotation::NotTested(false),
                 LineAnnotation::NotTested(false),
@@ -196,7 +227,7 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
                     path.to_str().unwrap(),
                     line_number
                 );
-                (region_annotation.clone(), region_annotation)
+                (region_annotation, region_annotation)
             }
 
             (LineMark::BeginMaybeTested, LineAnnotation::Tested(_)) => (
@@ -209,7 +240,20 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
                     path.to_str().unwrap(),
                     line_number
                 );
-                (region_annotation.clone(), region_annotation)
+                (region_annotation, region_annotation)
+            }
+
+            (LineMark::BeginFlakyTested, LineAnnotation::Tested(_)) => (
+                LineAnnotation::FlakyTested(false),
+                LineAnnotation::FlakyTested(false),
+            ),
+            (LineMark::BeginFlakyTested, region_annotation) => {
+                eprintln!(
+                    "{}:{}: ignored nested BEGIN FLAKY TESTED coverage annotation",
+                    path.to_str().unwrap(),
+                    line_number
+                );
+                (region_annotation, region_annotation)
             }
 
             (LineMark::EndNotTested, LineAnnotation::NotTested(_)) => (
@@ -222,7 +266,7 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
                     path.to_str().unwrap(),
                     line_number
                 );
-                (region_annotation.clone(), region_annotation)
+                (region_annotation, region_annotation)
             }
 
             (LineMark::EndMaybeTested, LineAnnotation::MaybeTested(_)) => (
@@ -235,40 +279,66 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
                     path.to_str().unwrap(),
                     line_number
                 );
-                (region_annotation.clone(), region_annotation)
+                (region_annotation, region_annotation)
             }
 
-            (LineMark::FileMaybeTested, region_annotation) => {
-                if is_file_not_tested || is_file_maybe_tested {
-                    eprintln!(
-                        "{}:{}: repeated FILE MAYBE TESTED coverage annotation",
-                        path.to_str().unwrap(),
-                        line_number
-                    );
-                }
-                is_file_maybe_tested = true;
-                (region_annotation.clone(), region_annotation)
+            (LineMark::EndFlakyTested, LineAnnotation::FlakyTested(_)) => (
+                LineAnnotation::FlakyTested(false),
+                LineAnnotation::Tested(false),
+            ),
+            (LineMark::EndFlakyTested, region_annotation) => {
+                eprintln!(
+                    "{}:{}: ignored nested END FLAKY TESTED coverage annotation",
+                    path.to_str().unwrap(),
+                    line_number
+                );
+                (region_annotation, region_annotation)
             }
 
             (LineMark::FileNotTested, region_annotation) => {
-                if is_file_not_tested || is_file_maybe_tested {
+                if is_file_not_tested || is_file_maybe_tested || is_file_flaky_tested {
                     eprintln!(
-                        "{}:{}: repeated FILE NOT TESTED coverage annotation",
+                        "{}:{}: repeated FILE NOT/MAYBE/FLAKY TESTED coverage annotation",
                         path.to_str().unwrap(),
                         line_number
                     );
                 }
                 is_file_not_tested = true;
-                (region_annotation.clone(), region_annotation)
+                (region_annotation, region_annotation)
+            }
+
+            (LineMark::FileMaybeTested, region_annotation) => {
+                if is_file_not_tested || is_file_maybe_tested || is_file_flaky_tested {
+                    eprintln!(
+                        "{}:{}: repeated FILE NOT/MAYBE/FLAKY TESTED coverage annotation",
+                        path.to_str().unwrap(),
+                        line_number
+                    );
+                }
+                is_file_maybe_tested = true;
+                (region_annotation, region_annotation)
+            }
+
+            (LineMark::FileFlakyTested, region_annotation) => {
+                if is_file_not_tested || is_file_maybe_tested || is_file_flaky_tested {
+                    eprintln!(
+                        "{}:{}: repeated FILE NOT/MAYBE/FLAKY TESTED coverage annotation",
+                        path.to_str().unwrap(),
+                        line_number
+                    );
+                }
+                is_file_flaky_tested = true;
+                (region_annotation, region_annotation)
             }
         };
         line_annotations.push(line_annotation);
         region_annotation = next_region_annotation;
     }
-    if is_file_maybe_tested {
+    if is_file_maybe_tested || (is_file_flaky_tested && flaky_policy == FlakyPolicy::MaybeTested) {
         verify_untested_file_annotations(path, &line_annotations);
         FileAnnotations::MaybeTested
-    } else if is_file_not_tested {
+    } else if is_file_not_tested || (is_file_flaky_tested && flaky_policy == FlakyPolicy::NotTested)
+    {
         verify_untested_file_annotations(path, &line_annotations);
         FileAnnotations::NotTested
     } else {
@@ -280,9 +350,9 @@ fn collect_file_annotations(path: &Path) -> FileAnnotations {
 fn verify_untested_file_annotations(path: &Path, line_annotations: &[LineAnnotation]) {
     for (mut line_number, line_annotation) in line_annotations.iter().enumerate() {
         line_number += 1;
-        if is_explicit(line_annotation) {
+        if is_explicit(*line_annotation) {
             eprintln!(
-                "{}:{}: line coverage annotation in a FILE which is NOT TESTED",
+                "{}:{}: line coverage annotation in a FILE which is NOT/MAYBE/FLAKY TESTED",
                 path.to_str().unwrap(),
                 line_number
             );
@@ -291,41 +361,47 @@ fn verify_untested_file_annotations(path: &Path, line_annotations: &[LineAnnotat
 }
 
 #[doc(hidden)]
-fn extract_line_mark(line: &str) -> LineMark {
-    if line.contains("// TESTED") || line.contains("/* TESTED") {
-        LineMark::LineTested
-    } else if line.contains("// MAYBE TESTED") || line.contains("/* MAYBE TESTED") {
-        LineMark::LineMaybeTested
-    } else if line.contains("// NOT TESTED")
-        || line.contains("/* NOT TESTED")
-        || line.contains("// APPEARS NOT TESTED")
+fn extract_line_mark(path: &str, line_number: usize, line: &str) -> LineMark {
+    if line.contains("// APPEARS NOT TESTED")
         || line.contains("/* APPEARS NOT TESTED")
-    {
-        LineMark::LineNotTested
-    } else if line.contains("// BEGIN MAYBE TESTED") || line.contains("/* BEGIN MAYBE TESTED") {
-        LineMark::BeginMaybeTested
-    } else if line.contains("// BEGIN NOT TESTED")
-        || line.contains("/* BEGIN NOT TESTED")
         || line.contains("// BEGIN APPEARS NOT TESTED")
         || line.contains("/* BEGIN APPEARS NOT TESTED")
-    {
-        LineMark::BeginNotTested
-    } else if line.contains("// END MAYBE TESTED") || line.contains("/* END MAYBE TESTED") {
-        LineMark::EndMaybeTested
-    } else if line.contains("// END NOT TESTED")
-        || line.contains("/* END NOT TESTED")
         || line.contains("// END APPEARS NOT TESTED")
         || line.contains("/* END APPEARS NOT TESTED")
-    {
-        LineMark::EndNotTested
-    } else if line.contains("// FILE MAYBE TESTED") || line.contains("// FILE MAYBE TESTED") {
-        LineMark::FileMaybeTested
-    } else if line.contains("// FILE NOT TESTED")
-        || line.contains("/* FILE NOT TESTED")
         || line.contains("// FILE APPEARS NOT TESTED")
         || line.contains("/* FILE APPEARS NOT TESTED")
     {
+        eprintln!(
+            "{}:{}: obsolete APPEARS TESTED directive, use FLAKY TESTED instead",
+            path, line_number
+        );
+        LineMark::None
+    } else if line.contains("// TESTED") || line.contains("/* TESTED") {
+        LineMark::LineTested
+    } else if line.contains("// MAYBE TESTED") || line.contains("/* MAYBE TESTED") {
+        LineMark::LineMaybeTested
+    } else if line.contains("// NOT TESTED") || line.contains("/* NOT TESTED") {
+        LineMark::LineNotTested
+    } else if line.contains("// FLAKY TESTED") || line.contains("/* FLAKY TESTED") {
+        LineMark::LineFlakyTested
+    } else if line.contains("// BEGIN MAYBE TESTED") || line.contains("/* BEGIN MAYBE TESTED") {
+        LineMark::BeginMaybeTested
+    } else if line.contains("// BEGIN NOT TESTED") || line.contains("/* BEGIN NOT TESTED") {
+        LineMark::BeginNotTested
+    } else if line.contains("// BEGIN FLAKY TESTED") || line.contains("/* BEGIN FLAKY TESTED") {
+        LineMark::BeginFlakyTested
+    } else if line.contains("// END MAYBE TESTED") || line.contains("/* END MAYBE TESTED") {
+        LineMark::EndMaybeTested
+    } else if line.contains("// END NOT TESTED") || line.contains("/* END NOT TESTED") {
+        LineMark::EndNotTested
+    } else if line.contains("// END FLAKY TESTED") || line.contains("/* END FLAKY TESTED") {
+        LineMark::EndFlakyTested
+    } else if line.contains("// FILE MAYBE TESTED") || line.contains("// FILE MAYBE TESTED") {
+        LineMark::FileMaybeTested
+    } else if line.contains("// FILE NOT TESTED") || line.contains("/* FILE NOT TESTED") {
         LineMark::FileNotTested
+    } else if line.contains("// FILE FLAKY TESTED") || line.contains("/* FILE FLAKY TESTED") {
+        LineMark::FileFlakyTested
     } else {
         LineMark::None
     }
@@ -413,6 +489,7 @@ fn canonical_file_name(sources: &[String], file_name: &str) -> Option<String> {
 
 #[doc(hidden)]
 fn report_wrong_annotations(
+    flaky_policy: FlakyPolicy,
     coverage_annotations: &HashMap<String, HashMap<i32, bool>>,
     source_annotations: &HashMap<String, FileAnnotations>,
 ) -> i32 {
@@ -428,6 +505,7 @@ fn report_wrong_annotations(
     for (file_name, coverage_line_annotations) in coverage_annotations {
         if (file_name.starts_with(src.as_str()) || file_name.starts_with(tests.as_str()))
             && report_file_wrong_annotations(
+                flaky_policy,
                 file_name,
                 coverage_line_annotations,
                 source_annotations.get(file_name).unwrap(),
@@ -449,6 +527,7 @@ fn report_wrong_annotations(
 
 #[doc(hidden)]
 fn report_file_wrong_annotations(
+    flaky_policy: FlakyPolicy,
     file_name: &str,
     coverage_file_annotations: &HashMap<i32, bool>,
     source_file_annotation: &FileAnnotations,
@@ -460,59 +539,65 @@ fn report_file_wrong_annotations(
             true
         }
         FileAnnotations::LineAnnotations(ref source_line_annotations) => {
-            let mut dis_report_annotation = false;
+            let mut did_report_annotation = false;
             for (mut line_number, source_line_annotation) in
                 source_line_annotations.iter().enumerate()
             {
                 line_number += 1;
                 let coverage_line_annotation = coverage_file_annotations.get(&(line_number as i32));
-                match (source_line_annotation, coverage_line_annotation) {
-                    (&LineAnnotation::Tested(_), Some(&false)) => {
+                match (
+                    flaky_policy,
+                    source_line_annotation,
+                    coverage_line_annotation,
+                ) {
+                    (_, &LineAnnotation::Tested(_), Some(&false))
+                    | (FlakyPolicy::Tested, &LineAnnotation::FlakyTested(_), Some(&true)) => {
                         eprintln!(
                             "{}:{}: wrong TESTED coverage annotation",
                             file_name, line_number,
                         );
-                        dis_report_annotation = true;
+                        did_report_annotation = true;
                     }
 
-                    (&LineAnnotation::NotTested(_), Some(&true)) => {
+                    (_, &LineAnnotation::NotTested(_), Some(&true))
+                    | (FlakyPolicy::NotTested, &LineAnnotation::FlakyTested(_), Some(&true)) => {
                         eprintln!(
                             "{}:{}: wrong NOT TESTED coverage annotation",
                             file_name, line_number,
                         );
-                        dis_report_annotation = true;
+                        did_report_annotation = true;
                     }
 
-                    (&LineAnnotation::Tested(true), None) => {
+                    (_, &LineAnnotation::Tested(true), None) => {
                         eprintln!(
                             "{}:{}: explicit TESTED coverage annotation for a non-executable line",
                             file_name, line_number,
                         );
-                        dis_report_annotation = true;
+                        did_report_annotation = true;
                     }
 
-                    (&LineAnnotation::NotTested(true), None) => {
+                    (_, &LineAnnotation::NotTested(true), None) => {
                         eprintln!(
                             "{}:{}: \
                              explicit NOT TESTED coverage annotation for a non-executable line",
                             file_name, line_number,
                         );
-                        dis_report_annotation = true;
+                        did_report_annotation = true;
                     }
 
-                    (&LineAnnotation::MaybeTested(true), None) => {
+                    (_, &LineAnnotation::MaybeTested(true), None) => {
                         eprintln!(
                             "{}:{}: \
                              explicit MAYBE TESTED coverage annotation for a non-executable line",
                             file_name, line_number,
                         );
-                        dis_report_annotation = true;
+                        did_report_annotation = true;
                     }
 
                     _ => {}
                 }
             }
-            dis_report_annotation
+            did_report_annotation
         }
     }
 }
@@ -532,44 +617,39 @@ fn report_uncovered_file_annotations(
 }
 
 #[doc(hidden)]
-fn process_args() {
-    let count = std::env::args().count();
+#[derive(Clone, Copy, PartialEq, Eq)]
+enum FlakyPolicy {
+    NotTested,
+    MaybeTested,
+    Tested,
+}
+
+#[doc(hidden)]
+fn process_args() -> FlakyPolicy {
+    let mut flaky_policy = FlakyPolicy::MaybeTested;
     let mut args = std::env::args();
-    let mut are_args_valid = true;
-    let mut should_print_version = false;
-
     args.next();
-    match count {
-        1 => {}
-        2 => match args.next().unwrap().as_ref() {
+    let program = args.next().unwrap();
+    for arg in args {
+        match arg.as_str() {
             "--version" => {
-                should_print_version = true;
+                println!("cargo-coverage-annotations {}", version!());
+                std::process::exit(0);
+            }
+            "--flaky=not-tested" => {
+                flaky_policy = FlakyPolicy::NotTested;
             }
-            "coverage-annotations" => {}
-            _ => {
-                are_args_valid = false;
+            "--flaky=maybe-tested" => {
+                flaky_policy = FlakyPolicy::MaybeTested;
             }
-        },
-        3 => {
-            if args.next().unwrap() == "coverage-annotations" && args.next().unwrap() == "--version"
-            {
-                should_print_version = true;
-            } else {
-                are_args_valid = false;
+            "--flaky=tested" => {
+                flaky_policy = FlakyPolicy::Tested;
             }
-        }
-        _ => {
-            are_args_valid = false;
+            arg => {
+                eprintln!("{}: unknown flag \"{}\"; valid flags are --version and --flaky=not-tested/maybe-tested/tested", program, arg);
+                std::process::exit(1);
+            }
         }
-    }
-
-    if !are_args_valid {
-        println!("cargo-coverage-annotations takes no arguments (except --version).");
-        std::process::exit(1);
-    }
-
-    if should_print_version {
-        println!("cargo-coverage-annotations {}", version!());
-        std::process::exit(0);
     }
+    flaky_policy
 }
